﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using SlimDX;
using SlimDX.Direct3D9;
using SlimDX.Windows;

namespace FDK
{
	public class CApplicationForm : IDisposable
	{
		// プロパティ

		public RenderForm Window
		{
			get;
			private set;
		}
		public Direct3D Direct3D
		{
			get;
			private set;
		}
		public CD3DSettings currentD3DSettings
		{
			get;
			private set;
		}
		public Device D3D9Device
		{
			get;
			private set;
		}


		// コンストラクタ

		public CApplicationForm()
		{
			this.Window = null;
			this.Direct3D = null;
			this.D3D9Device = null;
			this.currentD3DSettings = null;
		}


		// メソッド

		/// <summary>
		/// <para>メインループを実行する。</para>
		/// <para>ウィンドウが閉じられるか例外が発生すると、このメソッドを抜ける。</para>
		/// </summary>
		public void Run()
		{
			// ウィンドウを生成。

			#region [ SlimDX.Windows.RenderFrom 型のウィンドウを生成する。 ]
			//-----------------
			try
			{
				this.Window = Onウィンドウの作成();
			}
			catch
			{
				MessageBox.Show( "ウィンドウの生成に失敗しました。\n終了します。", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error );
				throw;
			}
			//-----------------
			#endregion
			#region [ ウィンドウイベントを設定する。 ]
			//-----------------
			this.Window.Closing += ( o, args ) => { this.bウィンドウが閉じられた = true; };
			//-----------------
			#endregion


			// Direct3D9 を生成。

			#region [ Direct3D9 を生成する。]
			//-----------------
			try
			{
				this.Direct3D = new Direct3D();
			}
			catch
			{
				MessageBox.Show( "Direct3D9 の生成に失敗しました。\n終了します。", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error );
				throw;
			}
			//-----------------
			#endregion
			#region [ アプリケーションを初期化し、Direct3D9 デバイスを生成する。]
			//-----------------
			this.On初期化();	// この中で this.Device も生成する。活性化（リソースの作成）も行う。

			if( this.D3D9Device == null )
				throw new InvalidOperationException( "アプリケーションにてDirect3D9デバイスが生成されていません。" );
			//-----------------
			#endregion


			// アプリケーションを起動。

			MessagePump.Run( this.Window, () => {

				// 終了チェック。

				if( this.bウィンドウが閉じられた )
					return;


				// 進行。

				this.On進行();


				// デバイスロストしてたらリセット。

				if( this.bデバイスロスト )
				{
					C共通.LogBlock( "Direct3Dデバイスをリセットする。あるいは、まだリセット不可なら 100ms スリープする。", () => {
						#region
						//-----------------
						if( this.D3D9Device.TestCooperativeLevel() == ResultCode.DeviceNotReset )
						{
							var newSettings = this.currentD3DSettings.Clone();

							#region [ ウィンドウモードの場合 → BackBuffer フォーマットと Desktop フォーマットを一致させる。]
							//-----------------
							if( this.currentD3DSettings.PresentParameters.Windowed )
							{
								DisplayMode dm = this.Direct3D.GetAdapterDisplayMode( this.currentD3DSettings.nAdaptor );
								if( this.currentD3DSettings.PresentParameters.BackBufferFormat != dm.Format )
								{
									C共通.LogBlock( "BackBuffer フォーマットが Desktopフォーマットと異なるので後者に合わせる", () => {

										Trace.TraceInformation( "BackBuffer フォーマット: " + this.currentD3DSettings.PresentParameters.BackBufferFormat );
										Trace.TraceInformation( "Desktop フォーマット: " + dm.Format );
										newSettings.PresentParameters.BackBufferFormat = dm.Format;

									} );
								}
							}
							//-----------------
							#endregion

							if( this.tDirect3Dデバイスを生成・変更・リセットする( newSettings ) )
							{
								Trace.TraceInformation( "Direct3Dデバイスを生成または変更しました。" );
								this.bデバイスロスト = false;
							}
						}
						else
						{
							Thread.Sleep( 100 );
						}
						//-----------------
						#endregion
					} );
				}


				// 描画。

				try
				{
					#region [ 描画を行う。]
					//-----------------
					this.D3D9Device.BeginScene();
					this.On描画();
					this.D3D9Device.EndScene();
					this.D3D9Device.Present();
					//-----------------
					#endregion
				}
				catch( Direct3D9Exception e )
				{
					#region [ デバイスロスト以外は全て throw する。]
					//-----------------
					if( e.ResultCode == ResultCode.DeviceLost )
					{
						Trace.TraceInformation( "Direct3Dデバイスを Lost しました。" );
						this.bデバイスロスト = true;
					}
					else
					{
						Trace.TraceInformation( "DeviceLost 以外の例外が発生しました。" );
						throw;
					}
					//-----------------
					#endregion
				}

			} );
			Trace.TraceInformation( "--------------------" );


			// アプリケーションを終了。

			C共通.LogBlock( "アプリケーションの終了", () => {

				this.On終了();		// ここで、非活性化（リソースの解放）も行う。
	
			} );
		}

		/// <summary>
		/// <para>Direct3Dデバイスの生成、変更、リセットを行う。</para>
		/// <para>新しい設定と現在の設定とを比較し、生成、変更、リセットのいずれかを実行する。</para>
		/// <para>処理に成功すれば true を返す。</para>
		/// </summary>
		public bool tDirect3Dデバイスを生成・変更・リセットする( CD3DSettings newD3DSettings )
		{
            C共通.LogBlock("Direct3Dデバイスの生成・変更・リセット", () =>
            {
                var oldD3DSettings = (this.currentD3DSettings != null) ? this.currentD3DSettings.Clone() : null;

                C共通.LogBlock("新しいモードに対応したウィンドウの設定", () =>
                {
                    #region
                    //-----------------
                    if (oldD3DSettings == null ||
                        !oldD3DSettings.PresentParameters.Windowed && newD3DSettings.PresentParameters.Windowed)	// 全画面→ウィンドウ時のみ（ウィンドウ→ウィンドウ時に二重設定してはいけない）
                    {
                        C共通.LogBlock("ウィンドウモードの設定", () =>
                        {
                            #region
                            //-----------------
                            C共通.LogBlock("ウィンドウモード用ウィンドウスタイルの設定", () => { CWin32.SetWindowLong(this.Window.Handle, CWin32.GWL_STYLE, this.wsウィンドウスタイル); });
                            //-----------------
                            #endregion
                        });
                    }
                    else if (oldD3DSettings == null ||
                        oldD3DSettings.PresentParameters.Windowed && !newD3DSettings.PresentParameters.Windowed)	// ウィンドウ→全画面時のみ（全画面→全画面時に二重設定してはいけない）
                    {
                        C共通.LogBlock("全画面モードの設定", () =>
                        {
                            #region
                            //-----------------
                            C共通.LogBlock("ウィンドウから全画面への切り替え（または最初から全画面）の場合はウィンドウパラメータを保存", () =>
                            {
                                #region
                                //-----------------
                                if (oldD3DSettings == null || oldD3DSettings.PresentParameters.Windowed)
                                {
                                    this.bTopMostWindow = this.Window.TopMost;
                                    this.wsウィンドウスタイル = CWin32.GetWindowLong(this.Window.Handle, CWin32.GWL_STYLE) & ~CWin32.WS_MAXIMIZE & ~CWin32.WS_MINIMIZE;
                                    this.WindowPlacement = new CWin32.WINDOWPLACEMENT() { length = CWin32.WINDOWPLACEMENT.Length };
                                    CWin32.GetWindowPlacement(this.Window.Handle, ref this.WindowPlacement);
                                }
                                //-----------------
                                #endregion
                            });
                            C共通.LogBlock("ウィンドウの不可視化", () => { this.Window.Hide(); });
                            C共通.LogBlock("全画面モード用ウィンドウスタイルの設定", () => { CWin32.SetWindowLong(this.Window.Handle, CWin32.GWL_STYLE, this.ws全画面スタイル); });
                            C共通.LogBlock("ウィンドウが最大化予定ならそれを拒否", () =>
                            {
                                #region
                                //-----------------
                                var pl = new CWin32.WINDOWPLACEMENT() { length = CWin32.WINDOWPLACEMENT.Length };
                                CWin32.GetWindowPlacement(this.Window.Handle, ref pl);
                                if ((pl.flags & CWin32.WPF_RESTORETOMAXIMIZED) != 0)
                                {
                                    pl.flags &= ~CWin32.WPF_RESTORETOMAXIMIZED;
                                    pl.showCmd = CWin32.EShowWindow.Restore;
                                    CWin32.SetWindowPlacement(this.Window.Handle, ref pl);
                                }
                                //-----------------
                                #endregion
                            });
                            //-----------------
                            #endregion
                        });
                    }
                    //-----------------
                    #endregion
                });

                C共通.LogBlock("Direct3D9デバイスのリセットまたは生成・再構築", () =>
                {
                    #region
                    //-----------------
                    bool bリセットでOK = (oldD3DSettings != null && oldD3DSettings.bデバイスの再生成が不要でリセットだけで済む(newD3DSettings));

                    if (bリセットでOK)
                    {
                        C共通.LogBlock("Direct3D9デバイスのリセット", () =>
                        {
                            #region
                            //-----------------
                            // リセットのときは Managedリソースのほうは解放しなくて良い。
                            C共通.LogBlock("Unmanageリソースの解放", () => { this.OnUnmanageリソースを解放する(); });
                            C共通.LogBlock("Direct3D9デバイスのリセット", () =>
                            {
                                #region
                                //-----------------
                                if (this.D3D9Device.Reset(newD3DSettings.PresentParameters) != ResultCode.DeviceLost)
                                {
                                    this.currentD3DSettings = newD3DSettings.Clone();
                                    Trace.TraceInformation("リセットに成功しました。");

                                    C共通.LogBlock("Unmanagedリソースの復元", () => { this.OnUnmanageリソースを生成する(); });
                                }
                                else
                                {
                                    Trace.TraceWarning("リセットに失敗しました。続けて再作成を試みます。");
                                    bリセットでOK = false;	// リセット失敗、再構成が必要。
                                }
                                //-----------------
                                #endregion
                            });
                            //-----------------
                            #endregion
                        });
                    }
                    if (!bリセットでOK)	// リセットに失敗した場合に bリセットでOK = false に設定してここに来るので、else は使わないこと。
                    {
                        C共通.LogBlock("デバイスの新規生成または再構築", () =>
                        {
                            #region
                            //-----------------
                            C共通.LogBlock("Managedリソースの解放", () => { this.OnManageリソースを解放する(); });
                            C共通.LogBlock("Unmanagedリソースの解放", () => { this.OnUnmanageリソースを解放する(); });
                            C共通.LogBlock("現在の Direct3D9デバイスの解放", () =>
                            {
                                #region
                                //-----------------
                                if (this.D3D9Device != null)
                                {
                                    this.D3D9Device.Dispose();
                                    this.D3D9Device = null;
                                    Trace.TraceInformation("Direct3D9デバイスを解放しました。");
                                }
                                else
                                    Trace.TraceInformation("Direct3D9デバイスが生成されていないため、何もしません。");
                                //-----------------
                                #endregion
                            });
                            C共通.LogBlock("Direct3D9デバイスの生成", () =>
                            {
                                #region
                                //-----------------
                                this.D3D9Device = new Device(		// 失敗したらそのまま例外をthrow
                                    this.Direct3D,
                                    newD3DSettings.nAdaptor,
                                    newD3DSettings.DeviceType,
                                    newD3DSettings.PresentParameters.DeviceWindowHandle,
                                    newD3DSettings.CreateFlags,
                                    newD3DSettings.PresentParameters);

                                Trace.TraceInformation("生成に成功しました。");
                                Trace.TraceInformation("Adapter:{0}, DeviceType:{1}, CreateFlags={2}",
                                    newD3DSettings.nAdaptor, newD3DSettings.DeviceType, newD3DSettings.CreateFlags);
                                Trace.TraceInformation("BackBuffer:{0}x{1} {2} {3}",
                                    newD3DSettings.PresentParameters.BackBufferWidth,
                                    newD3DSettings.PresentParameters.BackBufferHeight,
                                    newD3DSettings.PresentParameters.BackBufferFormat,
                                    (newD3DSettings.PresentParameters.Windowed ? "Windowed" : "Fullscreen"));

                                this.currentD3DSettings = newD3DSettings.Clone();
                                //-----------------
                                #endregion
                            });
                            C共通.LogBlock("Unmanagedリソースの復元", () => { this.OnUnmanageリソースを生成する(); });
                            C共通.LogBlock("Managedリソースの復元", () => { this.OnManageリソースを生成する(); });
                            //-----------------
                            #endregion
                        });
                    }

                    C共通.LogBlock("デバイスステータスの初期化", () => { this.OnD3Dデバイスステータスの初期化(); });
                    //-----------------
                    #endregion
                });

                C共通.LogBlock("全画面→ウィンドウ切替えの場合に保存していたウィンドウパラメータを復元", () =>
                {
                    #region
                    //-----------------
                    if (oldD3DSettings != null && !oldD3DSettings.PresentParameters.Windowed && newD3DSettings.PresentParameters.Windowed)
                    {
                        CWin32.SetWindowPlacement(this.Window.Handle, ref this.WindowPlacement);
                        this.Window.TopMost = this.bTopMostWindow;
                        Trace.TraceInformation("復元に成功しました。");
                    }
                    else
                        Trace.TraceInformation("全画面→ウインドウ切替えではないので何もしません。");
                    //-----------------
                    #endregion
                });

                C共通.LogBlock("ウィンドウを不可視にしてたら可視に戻す", () =>
                {
                    #region
                    //-----------------
                    if (!this.Window.Visible)
                    {
                        CWin32.ShowWindow(this.Window.Handle, CWin32.EShowWindow.Show);
                        Trace.TraceInformation("可視化しました。");
                    }
                    else
                        Trace.TraceInformation("不可視ではないので何もしません。");
                    //-----------------
                    #endregion
                });

                C共通.LogBlock("システムへ長期入力無反応時のアクションを指定", () =>
                {
                    #region
                    //-----------------
                    if (this.currentD3DSettings.PresentParameters.Windowed)
                    {
                        CWin32.SetThreadExecutionState(CWin32.ES_SYSTEM_REQUIRED | CWin32.ES_CONTINUOUS);
                        Trace.TraceInformation("ウィンドウモード時の指定を行いました。（システムは、入力反応が無ければスリープしたりモニタをOFFにしたりしてよい。）");
                    }
                    else
                    {
                        CWin32.SetThreadExecutionState(CWin32.ES_SYSTEM_REQUIRED | CWin32.ES_DISPLAY_REQUIRED | CWin32.ES_CONTINUOUS);
                        Trace.TraceInformation("全画面モード時の指定を行いました。（システムは、入力反応が無くてもスリープしたりモニタをいじったりしてはいけない。）");
                    }
                    //-----------------
                    #endregion
                });
            });

			return true;
		}

		
		// 子クラスで override するメソッド群

		/// <summary>
		/// <para>this.Window を生成する。</para>
		/// <para>作成に失敗したら null を返すこと。</para>
		/// </summary>
		protected virtual RenderForm Onウィンドウの作成()
		{
			/* 以下はサンプル。override するときの参考にすること。*/

			return new RenderForm( "Application Sample" );
		}

		/// <summary>
		/// <para>初期化、Direct3Dデバイスの生成、Managedリソースの生成の３つを行う。</para>
		/// <para>Window と Direct3D の生成の後に呼び出される。</para>
		/// <para>Direct3Dデバイスは、CApplicationForm.tDirect3Dデバイスの生成または変更( CD3DSettings newSettings ) を用いて行うこと。</para>
		/// </summary>
		protected virtual void On初期化() { }

		/// <summary>
		/// <para>Direct3Dデバイスに対する設定を行う。</para>
		/// <para>デバイスのリセット時や再作成時に呼び出される。</para>
		/// </summary>
		protected virtual void OnD3Dデバイスステータスの初期化()
		{
			// this.Device.SetTransform()
			// this.Device.SetRenderState()
			// this.Device.SetTextureStageState() 
			// etc,.
		}

		/// <summary>
		/// <para>進行処理を行う。</para>
		/// <para>BeginScene～EndScene の間に行えない処理もここで行う。</para>
		/// </summary>
		protected virtual void On進行() { }

		/// <summary>
		/// <para>描画処理を行う。</para>
		/// <para>BeginScene() と EndScene() の間に呼び出される。</para>
		/// <para>そのため、Direct3Dデバイスを変更するような操作は行わないこと。</para>
		/// </summary>
		protected virtual void On描画() { }

		/// <summary>
		/// <para>終了処理を行う。</para>
		/// <para>メインループを抜けた後、一番最後に呼び出される。</para>
		/// <para>Direct3D, Device は Dispose 不要。</para>
		/// </summary>
		protected virtual void On終了() { }

		/// <summary>
		/// <para>Managed リソースを生成する。</para>
		/// <para>デバイスを生成した直後に呼び出される。（リセットの直後には呼び出されない。）</para>
		/// </summary>
		protected virtual void OnManageリソースを生成する() { }

		/// <summary>
		/// <para>Managed リソースを解放する。</para>
		/// <para>デバイスを破棄する直前に呼び出される。（リセットの直前には呼び出されない。）</para>
		/// </summary>
		protected virtual void OnManageリソースを解放する() { }

		/// <summary>
		/// <para>Unmanaged リソースがあれば、それを生成する。</para>
		/// <para>デバイスロストした直後に呼び出される。</para>
		/// <para>連続して呼び出されても問題ないようにコーディングすること。</para>
		/// </summary>
		protected virtual void OnUnmanageリソースを生成する() { }

		/// <summary>
		/// <para>Unmanaged リソースがあれば、それを解放する。</para>
		/// <para>デバイスをリセットまたは破棄する直前に呼び出される。</para>
		/// </summary>
		protected virtual void OnUnmanageリソースを解放する() { }

		public void Dispose()
		{
			if( this.D3D9Device != null )
				this.D3D9Device.Dispose();

			if( this.Direct3D != null )
				this.Direct3D.Dispose();
		}

		#region [ private ]
		//-----------------
		bool bウィンドウが閉じられた = false;
		private bool bデバイスロスト = false;

		// ウィンドウ・全画面モード切替時の状態保存変数。

		private uint wsウィンドウスタイル = CWin32.WS_EX_APPWINDOW | CWin32.WS_EX_LAYERED | CWin32.WS_DLGFRAME | CWin32.WS_BORDER | CWin32.WS_CLIPCHILDREN | CWin32.WS_CLIPSIBLINGS;
		private uint ws全画面スタイル = CWin32.WS_POPUP | CWin32.WS_SYSMENU;
		private CWin32.WINDOWPLACEMENT WindowPlacement;
		private bool bTopMostWindow;
		//-----------------
		#endregion
	}
}
